#include "hourglass.h"

static void make_sched_param (enum thrd_pri_t pri, 
			      int *policy,
			      struct sched_param *param)
{
  int max_pri;

  switch (pri) {
  case NORMAL:
    param->sched_priority = 0;
    *policy = SCHED_OTHER;
    break;
    
  case MAIN_PRI:
  case TIMER_PRI:
  case HIGHEST:
    max_pri = sched_get_priority_max (SCHED_RR);
    if (max_pri == -1) {
      perror ("sched_get_priority_max");
      exit (errno);
    }
    if (pri == HIGHEST) {
      param->sched_priority = max_pri - 2;
    } else if (pri == TIMER_PRI) {
      param->sched_priority = max_pri - 1;
    } else {
      param->sched_priority = max_pri;
    }
    *policy = SCHED_RR;
    break;
    
  default:
    printf ("Hrm: unexpected arg to set_priority: %d\n", pri);
    exit (-1);
  }
}

void set_priority (enum thrd_pri_t pri)
{
  struct sched_param param;
  int policy, res;

  make_sched_param (pri, &policy, &param);

  res = sched_setscheduler (0, policy, &param);
  if (res != 0) {
    perror ("sched_setscheduler");
    exit (errno);
  }
}

void init_attr (pthread_attr_t *attr, 
		enum thrd_pri_t pri)
{
  struct sched_param param;
  int policy;
  
  make_sched_param (pri, &policy, &param);

  pthread_attr_init (attr);
  pthread_attr_setschedpolicy (attr, policy);
  pthread_attr_setschedparam (attr, &param);
}

void lock_pages (void)
{
  int res;
  char s1[] = "    **** WARNING: lack permissions to lock pages ****\n";
  char s2[] = "    **** results from this run may be unreliable ****\n";
  
  res = mlockall (MCL_CURRENT | MCL_FUTURE);
  if (res != 0) {
    perror ("mlockall");
    printf ("%s%s", s1, s2);
    fprintf (stderr, "%s%s", s1, s2);
  }
}

double find_mhz (void)
{
  /* various strategies are possible here; for Linux we just look in
     /proc, for Win32 we look at rate of change of system time
     vs. rdtsc; could also borrow LCM code from LMBench */
 
  int res;
  double freq;
  char str[MAXSTRLEN], *s2;
  FILE *f = fopen ("/proc/cpuinfo", "r");

  if (!f) {
    printf ("configuration error: could not open /proc/cpuinfo\n");
    printf ("in order to determine processor MHZ\n");
    exit (-1);
  }

  // uncool - this is fragile with respect to the format string...

  do {
    s2 = fgets (str, MAXSTRLEN, f);
    res = sscanf (str, "cpu MHz	: %lf", &freq);
  } while (res == 0 && s2);

  if (res == 0) {
    printf ("Hrm: failed to find MHz in /proc/cpuinfo\n");
    printf ("there was probably a configuration error\n");
    exit (-1);
  }

  return freq;
}

