#include "hourglass.h"

static void lat_test (int id)
{
  cycle_time now, expire;

  while (thrd_data[id].status != STOP) {
    expire = rdtsc() + us_to_cycle (5000.0);
    timed_sleep (id, -5000.0);
    now = rdtsc();
    
    if (thrd_data[id].lat_slot < LAT_SLOTS) {
      thrd_data[id].lat_slots[thrd_data[id].lat_slot] = now - expire;
      thrd_data[id].lat_slot++;
    }
  }
}

static int get_new_rec (void)
{
  long n;

  n = InterlockedIncrement (&current_record);

  if (n >= NUM_TRACE_RECORDS) {
    printf ("oops: trace record overrun - exiting\n");
    exit (-1);
  }

  return n;
}

#ifdef GAP_HIST
static void record_gap (cycle_time gap) 
{
  if (gap/GAP_HIST_DIVISOR > NUM_GAP_HIST_BUCKETS) {
    printf ("hrm: gap %lld exceeds NUM_GAP_HIST_BUCKETS %d\n",
	    gap/GAP_HIST_DIVISOR, NUM_GAP_HIST_BUCKETS);
    exit (-1);
  }
  gap_hist[gap/GAP_HIST_DIVISOR]++;
}
#endif

static cycle_time poll (int id)
{
  cycle_time now, gap, runtime;

  now = rdtsc ();
  gap = now - thrd_data[id].last_exec;

  if (gap > GAP) {
    if (thrd_data[id].exec_start == 0) {
      thrd_data[id].reserved_record = get_new_rec ();
    } else {
      int rec = thrd_data[id].reserved_record;
      thrd_data[id].reserved_record = get_new_rec ();
      trace_recs[rec].id = id;
      trace_recs[rec].start = 
	thrd_data[id].exec_start + 
	us_to_cycle (POLL_CORRECTION);
      trace_recs[rec].end = 
	thrd_data[id].last_exec - 
	us_to_cycle (POLL_CORRECTION);
    }
    thrd_data[id].exec_start = now;
    runtime = us_to_cycle(2*POLL_CORRECTION);
  } else {
    runtime = gap;
#ifdef GAP_HIST
    record_gap (gap);
#endif
  }

  thrd_data[id].last_exec = now;
  
  return runtime;
}

static void cpu_test (int id)
{
  while (thrd_data[id].status != STOP) {
    poll (id);
  }
}

static void cpu_yield_test (int id)
{
  cycle_time cpu_time;

  while (thrd_data[id].status != STOP) {
    cpu_time = 0;
    do {
      cpu_time += poll (id);
    } while (cpu_time <= us_to_cycle (500.0));
    sched_yield ();
  }
}

static void periodic_test (int id)
{
  cycle_time cpu_time;

  while (thrd_data[id].status != STOP) {
    timed_sleep (id, -5000.0);
    cpu_time = 0;
    do {
      cpu_time += poll (id);
    } while (cpu_time <= us_to_cycle (500.0));
  }
}

void *thrd_func (void *arg)
{
  int id = (int)arg;

  // set_priority (thrd_data[id].pri);

  /*
  thrd_data[id].status = READY;
  
  do { 
    sched_yield ();
    printf ("%d\n", id);
  } while (thrd_data[id].status != GO);
  */

  switch (thrd_data[id].task) {
    
  case LAT:
    lat_test (id);
    break;
    
  case PERIODIC:
    periodic_test (id);
    break;
    
  case YIELD:
    cpu_yield_test (id);
    break;
    
  case CPU:
    cpu_test (id);
    break;
    
  default:
    printf ("Hrm: thread %d has an unknown task\n", id);
    exit (-1);
  }

  thrd_data[id].status = DONE;

  return NULL;
}
